import shlex, subprocess, sys, time, traceback

from datetime import datetime
from izzylib import JsonConfig, DotDict, Path, boolean, logging
from threading import Thread
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler

from .config import config, path


def handle_shutdown(app):
	for proc in app.ctx.procs.values():
		if proc.active:
			proc.stop()

	## There seems to be a problem with python 3.10 or termux
	if sys.version_info >= (3, 10):
		try:
			loop = app.loop

		except:
			loop = asyncio.get_event_loop()

		loop.stop()
		loop.close()

		sys.exit()


class Processes(DotDict):
	def __init__(self, app):
		super().__init__()

		self._app = app
		self._app['procs'] = self

		#self._watcher = Observer()
		#self._watcher.schedule(WatchHandler(self), path.services, recursive=False)
		#self._watcher.start()

		self.scan()

		for proc in self.values():
			if proc.config.enabled:
				try:
					proc.start()
				except:
					traceback.print_exc()


	def scan(self):
		for filename in path.services.glob('*.json'):
			try:
				data = filename.json_load()
				name = data['name']

			except Exception as e:
				logging.error(f'Failed to load {filename.name}: {e.__class__.__name__}: {e}')
				continue

			self.load_service(name, filename)


	def load_service(self, name, filename):
		try:
			try:
				self[name].reload_config()

			except KeyError:
				self[name] = Process(filename)

			return True

		except Exception as e:
			logging.error('Error when loading service:', name)
			traceback.print_exc()


	def unload_service(self, name):
		try:
			config_path = self[name].config_path
			self[name].stop()
			del self[name]

			return config_path

		except KeyError:
			logging.verbose('Service not loaded:', name)


	def delete_service(self, name):
		config_path = self.unload_service(name)

		if config_path:
			config_path.delete()


	def get_service_by_path(self, filename):
		for proc in self.values():
			if filename == proc.config_path:
				return proc

		raise FileNotFoundError(f'Service file not found: {filename}')


class Process:
	config = {}


	def __init__(self, filename):
		self.config = JsonConfig(filename,
			name = 'noname',
			fullname = 'no name',
			description = 'No description',
			workdir = path.script,
			executable = None,
			arguments = [],
			enabled = True,
			autorestart = True,
			autorestart_timeout = 5,
			autorestart_tries = 5,
			terminate_timeout = 5,
			log_keep = 5
		)

		try:
			self.config.load()

		except FileNotFoundError:
			raise FileNotFoundError(f'Not a valid service: {self.name}')

		self.log = path.log.join(f'{self.name}.log')
		self.proc = None
		self._autostart = None
		self._stdout = None
		self._running = False

		self.executable = Path(self.executable)
		self.workdir = Path(self.workdir)


	def __getitem__(self, key):
		return self.config[key]


	def __setitem__(self, key, value):
		if key in ['executable', 'workdir']:
			value = Path(value)

		elif key in ['enabled', 'autorestart']:
			value = boolean(value)

		elif key in ['autorestart_timeout', 'autorestart_tries', 'terminate_timeout']:
			value = int(value)

		elif key == 'arguments' and type(value) == str:
			value = shlex.split(value)

		elif key in ['executable', 'workdir']:
			value = Path(value)

		self.config[key] = value


	def __getattr__(self, key):
		return self.config[key]


	def __setattr__(self, key, value):
		if key in self.config:
			return self.__setitem__(key, value)

		object.__setattr__(self, key, value)


	@property
	def active(self):
		if not self.proc:
			return False

		return True if self.proc.poll() == None else False


	@property
	def config_path(self):
		return self.config._json


	@config_path.setter
	def config_path(self, filename):
		self.config._json = Path(filename)


	@property
	def command(self):
		return ' '.join([self['executable'], *self['arguments']])


	def update_config(self, key, value):
		self[key] = value
		self.config.save()


	def open_log(self):
		if self._stdout:
			return

		self._stdout = self.log.open('a+')


	def close_log(self):
		if not self._stdout:
			return

		self._stdout.close()
		self._stdout = None


	def rotate_log(self):
		self.log.copy(path.logs.join(f'{self.name}-{date.isoformat()}.log'))
		self._stdout.truncate(0)


	def reload_config(self):
		logging.verbose('Reloading config:', self.name)
		return self.config.load()


	def start(self, first_start=True):
		if self.active:
			return logging.warning('Service already running:', self.name)

		if first_start:
			logging.info('Starting service:', self.fullname)

		self.open_log()

		self.proc = subprocess.Popen([self.executable.expanduser(), *self.arguments],
			cwd = self.workdir.expanduser(),
			stdout = self._stdout,
			stderr = self._stdout
		)

		self._running = True

		if self['autorestart'] and first_start:
			Thread(target=self.handle_restart).start()


	def stop(self):
		if not self.active:
			return logging.warning('Service is not running:', self.name)

		logging.info('Stopping service:', self.fullname)

		self.proc.terminate()

		try:
			self.proc.wait(self.terminate_timeout)

		except TimeoutExpired:
			self.proc.kill()

		if self.active:
			return logging.error('Failed to kill service:', self.name)

		self.close_log()
		self._running = False


	def restart(self):
		self.stop()
		self.start()


	def handle_restart(self):
		retries = 0

		while self._running:
			if not self.active:
				if self.autorestart_tries:
					retries += 1

				if retries <= self.autorestart_tries:
					time.sleep(self.autorestart_timeout)
					logging.info(f'{self.name}: Retry attempt {retries}')
					self.start(False)

				else:
					logging.warning(f'Failed to start service:', self['fullname'])
					self._running = False

			time.sleep(0.25)

		logging.verbose(f'{self.name}: Closing restart handler')


class WatchHandler(PatternMatchingEventHandler):
	def __init__(self, procs):
		super().__init__(
			patterns = ['*.json'],
			ignore_directories = True,
			case_sensitive = False
		)

		self.procs = procs


	def get_service_name(self, filename):
		data = filename.json_load()
		return data['name']


	def on_any_event(self, event):
		etype = event.event_type

		if etype not in ['modified', 'created', 'deleted', 'moved']:
			return

		src = Path(event.src_path)

		try:
			dest = Path(event.dest_path)
		except:
			dest = None

		if etype == 'moved':
			name = self.get_service_name(dest)

		else:
			name = self.get_service_name(src)

		if etype in ['modified', 'created']:
			self.procs.load_service(name, src)

		elif etype == 'deleted':
			proc = self.procs.get_service_by_path(src)
			self.procs.unload_service(proc.name)

		elif etype == 'moved':
			proc = self.procs.get_service_by_path(src)
			proc.config_path = dest
